@@ -37,6 +37,7 @@
 
 #include "ploop.h"
 #include "ploop_if.h"
+#include "cleanup.h"
 
 #define EXT4_IOC_OPEN_BALLOON		_IO('f', 42)
 
@@ -76,7 +77,6 @@ static int open_device(const char *device)
 	return fd;
 }
 
-
 static int fsync_balloon(int fd)
 {
 	if (fsync(fd)) {
@@ -96,7 +96,7 @@ int get_balloon(const char *mount_point, struct stat *st, int *outfd)
 	int fd, fd2;
 
 	if (mount_point == NULL)
-		return -1;
+		return SYSEXIT_PARAM;
 
 	fd = open(mount_point, O_RDONLY);
 	if (fd < 0) {
@@ -591,11 +591,12 @@ int ploop_balloon_complete(const char *device)
 
 	fd = open_device(device);
 	if (fd == -1)
-		return -1;
+		return SYSEXIT_OPEN;
 
-	err = ioctl(fd, PLOOP_IOC_DISCARD_FINI, NULL);
+	err = ioctl(fd, PLOOP_IOC_DISCARD_FINI);
 	if (err && errno != EBUSY) {
 		ploop_err(errno, "Can't finalize discard mode");
+		err = SYSEXIT_DEVIOC;
 		goto out;
 	}
 
@@ -817,6 +818,11 @@ static void stop_trim_handler(int sig)
 	trim_stop = 1;
 }
 
+static void cancel_discard(void *data)
+{
+	ploop_balloon_complete(data);
+}
+
 /* The fragmentation of such blocks doesn't affect the speed of w/r */
 #define MAX_DISCARD_CLU 32
 
@@ -870,16 +876,47 @@ static int ploop_trim(const char *mount_point, __u64 minlen_b, __u64 cluster)
 	return ret;
 }
 
+static int blk_discard(int fd, __u32 cluster, __u64 start, __u64 len)
+{
+	__u64 max_discard_len = S2B(B2S(UINT_MAX) / cluster * cluster);
+
+	while (len > 0) {
+		__u64 range[2];
+		int ret;
+
+		range[0] = start;
+		range[1] = MIN(len, max_discard_len);
+
+		if (start % S2B(cluster) && len > range[1])
+			range[1] -= start % S2B(cluster);
+
+		ploop_log(1, "Call BLKDISCARD start=%llu length=%llu ", range[0], range[1]);
+		ret = ioctl_device(fd, BLKDISCARD, range);
+		if (ret)
+			return ret;
+
+		start += range[1];
+		len -= range[1];
+	}
+
+	return 0;
+}
+
 static int __ploop_discard(struct ploop_disk_images_data *di, int fd,
 			const char *device, const char *mount_point,
 			__u64 minlen_b, __u32 cluster, __u32 to_free,
-			const int *stop)
+			__u64 blk_discard_range[2], const int *stop)
 {
 	pid_t tpid;
 	int err = 0, ret, status;
 	__u32 size = 0;
+	struct ploop_cleanup_hook *h;
 
-	ploop_log(3, "Trying to find free extents bigger than %llu bytes", minlen_b);
+	if (blk_discard_range != NULL)
+		ploop_log(0, "Discard %s start=%llu length=%llu",
+				device, blk_discard_range[0], blk_discard_range[1]);
+	else
+		ploop_log(3, "Trying to find free extents bigger than %llu bytes", minlen_b);
 
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
@@ -893,18 +930,24 @@ static int __ploop_discard(struct ploop_disk_images_data *di, int fd,
 	tpid = fork();
 	if (tpid < 0) {
 		ploop_err(errno, "Can't fork");
-		if (ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL))
+		ret = ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL);
+		if (ret) {
 			ploop_err(errno, "Can't finalize discard mode");
-
-		return -1;
+			return ret;
+		}
 	}
 
+	h = register_cleanup_hook(cancel_discard, (void *) device);
+
 	if (tpid == 0) {
-		ret = ploop_trim(mount_point, minlen_b, cluster);
+		if (blk_discard_range != NULL)
+			ret = blk_discard(fd, cluster, blk_discard_range[0], blk_discard_range[1]);
+		else
+			ret = ploop_trim(mount_point, minlen_b, cluster);
 		if (ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL))
 			ploop_err(errno, "Can't finalize discard mode");
 
-		exit(ret < 0);
+		exit(ret != 0);
 	}
 
 	while (1) {
@@ -970,7 +1013,6 @@ static int __ploop_discard(struct ploop_disk_images_data *di, int fd,
 			break;
 	}
 
-
 	if (ret) {
 		err = -1;
 
@@ -988,9 +1030,14 @@ static int __ploop_discard(struct ploop_disk_images_data *di, int fd,
 		ploop_log(0, "%d clusters have been relocated", size);
 	}
 
-	ret = waitpid(tpid, &status, 0);
+	unregister_cleanup_hook(h);
+
+	while ((ret = waitpid(tpid, &status, 0)))
+		 if (errno != EINTR)
+			break;
 	if (ret == -1) {
-		ploop_err(errno, "wait() failed");
+		if (errno != ECHILD)
+			ploop_err(errno, "wait() failed");
 		err = -1;
 	} else if (WIFEXITED(status)) {
 		ret = WEXITSTATUS(status);
@@ -1015,12 +1062,13 @@ static int do_ploop_discard(struct ploop_disk_images_data *di,
 {
 	int fd, ret;
 	int blocksize;
-	__u32  cluster;
+	__u32 cluster;
 
 	if (ploop_get_attr(device, "block_size", &blocksize)) {
 		ploop_err(0, "Can't find block size");
 		return SYSEXIT_SYSFS;
 	}
+
 	cluster = S2B(blocksize);
 
 	if (to_free == 0)
@@ -1028,7 +1076,7 @@ static int do_ploop_discard(struct ploop_disk_images_data *di,
 	to_free = to_free / cluster;
 	if (!to_free) {
 		ploop_err(0, "Can't shrink by less than %d bytes", cluster);
-		return -1;
+		return SYSEXIT_PARAM;
 	}
 
 	fd = open_device(device);
@@ -1036,7 +1084,35 @@ static int do_ploop_discard(struct ploop_disk_images_data *di,
 		return SYSEXIT_OPEN;
 
 	ret = __ploop_discard(di, fd, device, mount_point,
-					minlen_b, cluster, to_free, stop);
+					minlen_b, cluster, to_free, NULL, stop);
+
+	close(fd);
+
+	return ret;
+}
+
+int ploop_blk_discard(const char* device, __u32 blocksize, off_t start, off_t end)
+{
+	int ret, fd;
+	__u64 range[2];
+
+	start = S2B(start);
+	end = S2B(end);
+
+	if (start >= end)
+		return 0;
+
+	range[0] = start;
+	range[1] = end - start;
+
+	fd = open(device, O_RDWR);
+	if (fd < 0) {
+		ploop_err(errno, "Can't open ploop device %s",
+				device);
+		return SYSEXIT_OPEN;
+	}
+
+	ret = __ploop_discard(NULL, fd, device, NULL, 0, S2B(blocksize), ~0U, range, NULL);
 
 	close(fd);
 
@@ -1147,29 +1223,29 @@ int ploop_complete_running_operation(const char *device)
 		goto err;
 
 	ploop_log(0, "Completing an on-going operation %s for device %s",
-			mntn2str(b_ctl.mntn_type), device);
+		mntn2str(b_ctl.mntn_type), device);
 
 	switch (b_ctl.mntn_type) {
-		case PLOOP_MNTN_MERGE:
-			ret = ioctl_device(fd, PLOOP_IOC_MERGE, 0);
-			break;
-		case PLOOP_MNTN_GROW:
-			ret = ioctl_device(fd, PLOOP_IOC_GROW, 0);
-			break;
-		case PLOOP_MNTN_RELOC:
-		case PLOOP_MNTN_FBLOADED:
-			ret = ploop_balloon_complete(device);
-			break;
-		case PLOOP_MNTN_TRACK:
-			ret = ioctl_device(fd, PLOOP_IOC_TRACK_ABORT, 0);
-			break;
-		case PLOOP_MNTN_DISCARD:
-			ret = ploop_balloon_complete(device);
-			break;
-		case PLOOP_MNTN_BALLOON:
-			/*  FIXME : ploop_balloon_check_and_repair(device, mount_point, 1; */
-			ret = 0;
-			break;
+	case PLOOP_MNTN_MERGE:
+		ret = ioctl_device(fd, PLOOP_IOC_MERGE, 0);
+		break;
+	case PLOOP_MNTN_GROW:
+		ret = ioctl_device(fd, PLOOP_IOC_GROW, 0);
+		break;
+	case PLOOP_MNTN_RELOC:
+	case PLOOP_MNTN_FBLOADED:
+		ret = ploop_balloon_complete(device);
+		break;
+	case PLOOP_MNTN_TRACK:
+		ret = ioctl_device(fd, PLOOP_IOC_TRACK_ABORT, 0);
+		break;
+	case PLOOP_MNTN_DISCARD:
+		ret = ploop_balloon_complete(device);
+		break;
+	case PLOOP_MNTN_BALLOON:
+		/*  FIXME : ploop_balloon_check_and_repair(device, mount_point, 1; */
+		ret = 0;
+		break;
 	}
 
 err:
