@@ -133,7 +133,7 @@ static int nread(int fd, void * buf, int len)
 	return -1;
 }
 
-int receive_process(const char *dst)
+int ploop_receive(const char *dst)
 {
 	int ofd, ret;
 	__u64 cluster = 0;
@@ -283,14 +283,14 @@ static int open_mount_point(const char *device)
 	return fd;
 }
 
-int send_process(const char *device, int ofd, const char *flush_cmd,
+int ploop_send(const char *device, int ofd, const char *flush_cmd,
 		int is_pipe)
 {
 	struct delta idelta = { .fd = -1 };
 	int tracker_on = 0;
 	int fs_frozen = 0;
 	int devfd = -1;
-	int mntfd;
+	int mntfd = -1;
 	int ret = 0;
 	char *send_from = NULL;
 	char *format = NULL;
@@ -317,7 +317,7 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 	}
 
 	mntfd = open_mount_point(device);
-	if (devfd < 0) {
+	if (mntfd < 0) {
 		/* Error is printed by open_mount_point() */
 		ret = SYSEXIT_OPEN;
 		goto done;
@@ -334,19 +334,13 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 		goto done;
 	}
 
-	iter = 0;
-	while (ioctl(devfd, PLOOP_IOC_TRACK_INIT, &e)) {
-		if (errno == EBUSY && iter++ == 0) {
-			/* Previous instance was killed?
-			 * Abort tracking and retry
-			 */
-			ioctl(devfd, PLOOP_IOC_TRACK_ABORT, 0);
-			continue;
-		}
-		ploop_err(errno, "PLOOP_IOC_TRACK_INIT");
-		ret = SYSEXIT_DEVIOC;
+	ret = ploop_complete_running_operation(device);
+	if (ret)
+		goto done;
+
+	ret = ioctl_device(devfd, PLOOP_IOC_TRACK_INIT, &e);
+	if (ret)
 		goto done;
-	}
 	tracker_on = 1;
 
 	if (open_delta_simple(&idelta, send_from, O_RDONLY|O_DIRECT, OD_NOFLAGS)) {
@@ -361,11 +355,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 		int n;
 
 		trackpos = pos + cluster;
-		if (ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos)) {
-			ploop_err(errno, "PLOOP_IOC_TRACK_SETPOS");
-			ret = SYSEXIT_DEVIOC;
+		ret = ioctl_device(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos);
+		if (ret)
 			goto done;
-		}
 
 		n = idelta.fops->pread(idelta.fd, iobuf, cluster, pos);
 		if (n < 0) {
@@ -471,18 +463,14 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 	}
 
 	/* Flush journal and freeze fs (this also clears the fs dirty bit) */
-	if (ioctl(mntfd, FIFREEZE, 0)) {
-		ploop_err(errno, "ioctl(FIFREEZE) failed");
-		ret = SYSEXIT_FSYNC;
+	ret = ioctl_device(mntfd, FIFREEZE, 0);
+	if (ret)
 		goto done;
-	}
 	fs_frozen = 1;
 
-	if (ioctl(devfd, PLOOP_IOC_SYNC, 0)) {
-		ploop_err(errno, "PLOOP_IOC_SYNC");
-		ret = SYSEXIT_DEVIOC;
+	ret = ioctl(devfd, PLOOP_IOC_SYNC, 0);
+	if (ret)
 		goto done;
-	}
 
 	iter = 1;
 	iterpos = 0;
@@ -511,11 +499,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 					copy = cluster;
 				if (pos + copy > trackpos) {
 					trackpos = pos + copy;
-					if (ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos)) {
-						ploop_err(errno, "PLOOP_IOC_TRACK_SETPOS");
-						ret = SYSEXIT_DEVIOC;
+					ret = ioctl(devfd, PLOOP_IOC_TRACK_SETPOS, &trackpos);
+					if (ret)
 						goto done;
-					}
 				}
 				n = idelta.fops->pread(idelta.fd, iobuf, copy, pos);
 				if (n < 0) {
@@ -572,11 +558,9 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 		}
 	}
 
-	if (ioctl(devfd, PLOOP_IOC_TRACK_STOP, 0)) {
-		ploop_err(errno, "PLOOP_IOC_TRACK_STOP");
-		ret = SYSEXIT_DEVIOC;
+	ret = ioctl(devfd, PLOOP_IOC_TRACK_STOP, 0);
+	if (ret)
 		goto done;
-	}
 	tracker_on = 0;
 
 	ret = send_buf(ofd, iobuf, 0, 0, is_pipe);
@@ -586,18 +570,16 @@ int send_process(const char *device, int ofd, const char *flush_cmd,
 	}
 
 done:
-	if (fs_frozen) {
-		if (ioctl(mntfd, FITHAW, 0))
-			ploop_err(errno, "ioctl(FITHAW) failed");
-	}
-	if (tracker_on) {
-		if (ioctl(devfd, PLOOP_IOC_TRACK_ABORT, 0))
-			ploop_err(errno, "PLOOP_IOC_TRACK_ABORT");
-	}
+	if (fs_frozen)
+		(void)ioctl_device(mntfd, FITHAW, 0);
+	if (tracker_on)
+		(void)ioctl_device(devfd, PLOOP_IOC_TRACK_ABORT, 0);
 	if (iobuf)
 		free(iobuf);
 	if (devfd >=0)
 		close(devfd);
+	if (mntfd >=0)
+		close(mntfd);
 	if (send_from)
 		free(send_from);
 	if (idelta.fd >= 0)
