@@ -33,6 +33,7 @@
 #include <sys/vfs.h>
 #include <sys/syscall.h>
 #include <mntent.h>
+#include <ext2fs/ext2_fs.h>
 
 #include "ploop.h"
 #include "cleanup.h"
@@ -235,7 +236,7 @@ static int get_temp_mountpoint(const char *file, int create, char *buf, int len)
 	return 0;
 }
 
-static int check_blockdev_size(unsigned long long sectors, __u32 blocksize)
+static int check_size(unsigned long long sectors)
 {
 	const unsigned long long max = (__u32)-1;
 
@@ -245,6 +246,14 @@ static int check_blockdev_size(unsigned long long sectors, __u32 blocksize)
 				sectors, max);
 		return -1;
 	}
+	return 0;
+}
+
+static int check_blockdev_size(unsigned long long sectors, __u32 blocksize)
+{
+	if (check_size(sectors))
+		return -1;
+
 	if (sectors % blocksize) {
 		ploop_err(0, "An incorrect block device size is specified: %llu sectors."
 				" The block device size must be aligned to the cluster block size %d.",
@@ -368,7 +377,7 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 				l2_slot < vh.m_Size;
 				i++, l2_slot++)
 		{
-			off = vh.m_FirstBlockOffset + (l2_slot * blocksize);
+			off = (off_t)vh.m_FirstBlockOffset + (l2_slot * blocksize);
 			((__u32*)buf)[i] = off;
 		}
 		if (WRITE(odelta.fd, buf, cluster))
@@ -474,8 +483,16 @@ static int create_image(struct ploop_disk_images_data *di,
 	char fname[PATH_MAX];
 	struct stat st;
 
-	if (size_sec == 0 || file == NULL)
+	if (size_sec == 0) {
+		ploop_err(0, "Incorrect block device size specified: "
+				"%lu sectors", (long)size_sec);
 		return SYSEXIT_PARAM;
+	}
+	if (file == NULL) {
+		ploop_err(0, "Image file name not specified");
+		return SYSEXIT_PARAM;
+	}
+
 	if (stat(file, &st) == 0) {
 		ploop_err(0, "File already exists %s", file);
 		return SYSEXIT_PARAM;
@@ -590,9 +607,10 @@ static int ploop_init_image(struct ploop_disk_images_data *di, struct ploop_crea
 		goto err;
 
 err:
-	if (ploop_umount_image(di))
+	if (ploop_umount_image(di)) {
 		if (ret == 0)
 			ret = SYSEXIT_UMOUNT;
+	}
 
 	return ret;
 }
@@ -628,6 +646,9 @@ int ploop_create_image(struct ploop_create_param *param)
 	int ret;
 	__u32 blocksize;
 
+	if (check_size(param->size))
+		return SYSEXIT_PARAM;
+
 	blocksize = param->blocksize ?
 			param->blocksize : (1 << PLOOP1_DEF_CLUSTER_LOG);
 	if (!is_valid_blocksize(blocksize)) {
@@ -641,15 +662,18 @@ int ploop_create_image(struct ploop_create_param *param)
 		return SYSEXIT_NOMEM;
 	di->blocksize = blocksize;
 	ret = create_image(di, param->image, di->blocksize,
-			ROUNDUP_BDSIZE(param->size, di->blocksize),
+			ROUND_BDSIZE(param->size, di->blocksize),
 			param->mode);
 	if (ret)
-		return ret;
+		goto out;
+
 	if (param->fstype != NULL) {
 		ret = ploop_init_image(di, param);
 		if (ret)
 			ploop_drop_image(di);
 	}
+
+out:
 	ploop_free_diskdescriptor(di);
 
 	return ret;
@@ -988,7 +1012,7 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 	 * 1 mount ro and read balloon inode
 	 * 2 remount with balloon_ino=ino
 	 */
-	if (mount(part_device, param->target,	fstype, MS_RDONLY, NULL)) {
+	if (mount(part_device, param->target, fstype, MS_RDONLY, param->mount_data)) {
 		ploop_err(errno, "Can't mount file system dev=%s target=%s",
 				part_device, param->target);
 		return SYSEXIT_MOUNT;
@@ -1017,8 +1041,7 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 	return 0;
 }
 
-static int add_delta(int lfd, const char *component_name,
-		const char *image, struct ploop_ctl_delta *req, int level)
+static int add_delta(int lfd, const char *image, struct ploop_ctl_delta *req)
 {
 	int fd;
 	int ro = (req->c.pctl_flags & PLOOP_FMT_RDONLY);
@@ -1032,23 +1055,6 @@ static int add_delta(int lfd, const char *component_name,
 
 	req->f.pctl_fd = fd;
 
-	if (level == 0) {
-		/* Base image: check if in use */
-		char dev[64];
-		int rc;
-
-		ret = SYSEXIT_PARAM;
-		rc = ploop_find_dev(component_name, image, dev, sizeof(dev));
-		if (rc < 0)
-			goto out;
-		if (rc == 0) {
-			ret = SYSEXIT_MOUNT;
-			ploop_err(0, "Image %s already used by device %s",
-					image, dev);
-			goto out;
-		}
-	}
-
 	if (ioctl(lfd, PLOOP_IOC_ADD_DELTA, req) < 0) {
 		ploop_err(0, "Can't add image %s: %s", image,
 				(errno == ENOTSUP) ?
@@ -1101,7 +1107,6 @@ static int add_deltas(struct ploop_disk_images_data *di,
 	int i;
 	int ret = 0;
 	struct ploop_ctl_delta req = {};
-	const char *component_name = di ? di->runtime->component_name : NULL;
 
 	if (device[0] == '\0') {
 		char buf[64];
@@ -1151,7 +1156,7 @@ static int add_deltas(struct ploop_disk_images_data *di,
 
 		ploop_log(0, "Adding delta dev=%s img=%s (%s)",
 				device, image, ro ? "ro" : "rw");
-		ret = add_delta(*lfd_p, component_name, image, &req, i);
+		ret = add_delta(*lfd_p, image, &req);
 		if (ret)
 			goto err1;
 	}
@@ -1182,6 +1187,52 @@ err:
 	return ret;
 }
 
+#ifndef EXT4_SUPER_MAGIC
+#define EXT4_SUPER_MAGIC	0xEF53
+#endif
+#ifndef FS_IOC_GETFLAGS
+#define FS_IOC_GETFLAGS	_IOR('f', 1, long)
+#endif
+#ifndef EXT4_EXTENTS_FL
+#define EXT4_EXTENTS_FL		0x00080000 /* Inode uses extents */
+#endif
+
+static int check_mount_restrictions(struct ploop_mount_param *param, const char *fname)
+{
+	struct statfs st;
+	int fd;
+	long flags;
+
+	/* FIXME: */
+	if (getenv("PLOOP_SKIP_EXT4_EXTENTS_CHECK") != NULL)
+		return 0;
+	if (statfs(fname, &st) < 0) {
+		ploop_err(errno, "Unable to statfs %s", fname);
+		return -1;
+	}
+	if (st.f_type == EXT4_SUPER_MAGIC) {
+		fd = open(fname, O_RDONLY);
+		if (fd < 0) {
+			ploop_err(errno, "Can't open %s", fname);
+			return -1;
+		}
+		if (ioctl(fd, FS_IOC_GETFLAGS, &flags) < 0) {
+			ploop_err(errno, "FS_IOC_GETFLAGS %s", fname);
+			close(fd);
+			return -1;
+		}
+		close(fd);
+
+		if (!(flags & EXT4_EXTENTS_FL)) {
+			ploop_err(0, "The ploop image can not be used on ext3 or ext4 file"
+					" system without extents.");
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 int ploop_mount(struct ploop_disk_images_data *di, char **images,
 		struct ploop_mount_param *param, int raw)
 {
@@ -1219,6 +1270,9 @@ int ploop_mount(struct ploop_disk_images_data *di, char **images,
 	} else if (di)
 		blocksize = di->blocksize;
 
+	if (check_mount_restrictions(param, images[0]))
+		return SYSEXIT_MOUNT;
+
 	for (i = 0; images[i] != NULL; i++) {
 		int ro;
 		int flags = FSCK_CHECK | (di ? FSCK_DROPINUSE : 0);
@@ -1245,7 +1299,6 @@ int ploop_mount(struct ploop_disk_images_data *di, char **images,
 	}
 
 	ret = add_deltas(di, images, param, raw, blocksize, &lfd);
-
 	if (ret)
 		goto err;
 
@@ -1295,7 +1348,9 @@ static int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_par
 	images = make_images_list(di, guid, 0);
 	if (images == NULL)
 		return SYSEXIT_NOMEM;
+
 	ret = ploop_mount(di, images, param, (di->mode == PLOOP_RAW_MODE));
+
 	free_images_list(images);
 
 	return ret;
@@ -1413,12 +1468,12 @@ int ploop_umount_image(struct ploop_disk_images_data *di)
 	ret = ploop_find_dev_by_uuid(di, 0, dev, sizeof(dev));
 	if (ret == -1) {
 		ploop_unlock_di(di);
-		return -1;
+		return SYSEXIT_SYS;
 	}
 	if (ret != 0) {
 		ploop_unlock_di(di);
 		ploop_err(0, "Image %s is not mounted", di->images[0]->file);
-		return -1;
+		return SYSEXIT_DEV_NOT_MOUNTED;
 	}
 
 	ret = ploop_complete_running_operation(dev);
@@ -1476,39 +1531,141 @@ int ploop_grow_device(const char *device, __u32 blocksize, off_t new_size)
 	return 0;
 }
 
+static int ploop_raw_discard(struct ploop_disk_images_data *di, const char *device,
+		__u32 blocksize, off_t start, off_t end)
+{
+	int ret;
+	char conf[PATH_MAX];
+	off_t new_end;
+
+	new_end = start + GPT_DATA_SIZE;
+	new_end = ROUNDUP_BDSIZE(new_end, blocksize);
+
+	if (new_end >= end)
+		return 0;
+
+	ret = resize_gpt_partition(device, new_end);
+	if (ret)
+		return ret;
+
+	ret = ploop_stop_device(device);
+	if (ret)
+		return ret;
+
+	ploop_log(0, "Truncate %s %lu",
+			di->images[0]->file, (long)S2B(new_end));
+	if (truncate(di->images[0]->file, S2B(new_end))) {
+		ploop_err(errno, "Failed to truncate %s",
+				di->images[0]->file);
+		return SYSEXIT_FTRUNCATE;
+	}
+
+	di->size = new_end;
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	ret = ploop_store_diskdescriptor(conf, di);
+	if (ret)
+		return SYSEXIT_DISKDESCR;
+
+	return 0;
+}
+
+/* The code below works correctly only if
+ *	device=/dev/ploopN
+ *	part_dev_size=/dev/ploopNp1
+ */
+static int shrink_device(struct ploop_disk_images_data *di,
+		const char *device, const char *part_device,
+		off_t part_dev_size, off_t new_size, __u32 blocksize)
+{
+	struct dump2fs_data data;
+	char buf[PATH_MAX];
+	__u32 part_start;
+	int ret;
+	char *p1, *p2;
+	int top, raw;
+	off_t start, end;
+
+	p1 = strrchr(device, '/');
+	p2 = strrchr(part_device, '/');
+	snprintf(buf, sizeof(buf), "/sys/block/%s/%s/start", p1, p2);
+	if (get_dev_start(buf, &part_start)) {
+		ploop_err(0, "Can't find out offset from start of ploop device (%s)",
+				part_device);
+		return SYSEXIT_SYSFS;
+	}
+	ret = ploop_get_attr(device, "top", &top);
+	if (ret)
+		return SYSEXIT_SYSFS;
+
+	raw = (di->mode == PLOOP_RAW_MODE && top == 0);
+	ploop_log(0, "Offline shrink %s dev=%s size=%lu new_size=%lu, start=%u",
+			(raw) ? "raw" : "",
+			part_device, (long)part_dev_size, (long)new_size, part_start);
+	ret = e2fsck(part_device);
+	if (ret)
+		return ret;
+
+	/* offline resize */
+	ret = resize_fs(part_device, new_size);
+	if (ret)
+		return ret;
+
+	ret = dumpe2fs(part_device, &data);
+	if (ret)
+		return ret;
+
+	start = part_start + B2S(data.block_count * data.block_size);
+	end = part_start + part_dev_size;
+	if (raw)
+		ret = ploop_raw_discard(di, device, blocksize, start, end);
+	else
+		ret = ploop_blk_discard(device, blocksize, start, end);
+
+	if (ret)
+		return ret;
+
+	return 0;
+}
+
 int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_param *param)
 {
 	int ret;
 	struct ploop_mount_param mount_param = {};
 	char buf[PATH_MAX];
-	int mounted = 0;
+	char part_device[64];
+	int mounted = -1;
 	int balloonfd = -1;
 	struct stat st;
+	off_t part_dev_size = 0;
 	off_t dev_size = 0;
 	__u64 balloon_size = 0;
 	__u64 new_balloon_size = 0;
 	struct statfs fs;
-	unsigned long long new_size = ROUNDUP_BDSIZE(param->size, di->blocksize);
+	unsigned long long new_size;
+	__u32 blocksize = 0;
+	off_t new_fs_size = 0;
 
 	if (di->nimages == 0) {
 		ploop_err(0, "No images in DiskDescriptor");
-		return -1;
+		return SYSEXIT_DISKDESCR;
 	}
 
-	if (check_blockdev_size(new_size, di->blocksize))
-		return -1;
+	if (check_size(param->size))
+		return SYSEXIT_PARAM;
 
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
 	ret = ploop_find_dev_by_uuid(di, 1, buf, sizeof(buf));
-	if (ret == -1)
+	if (ret == -1) {
+		ret = SYSEXIT_SYS;
 		goto err;
+	}
 	if (ret != 0) {
 		ret = auto_mount_image(di, &mount_param);
 		if (ret)
 			goto err;
-		mounted = 1;
+		mounted = 0;
 	} else {
 		ret = ploop_complete_running_operation(buf);
 		if (ret)
@@ -1521,11 +1678,40 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			goto err;
 		}
 		mount_param.target = strdup(buf);
+		mounted = 1;
+	}
+
+	/* FIXME: Deny resize image if there are children */
+
+	ret = SYSEXIT_SYSFS;
+	if (ploop_get_attr(mount_param.device, "block_size", (int *)&blocksize))
+		goto err;
+
+	new_size = ROUND_BDSIZE(param->size, blocksize);
+
+	ret = get_partition_device_name(mount_param.device, part_device, sizeof(part_device));
+	if (ret) {
+		ret = SYSEXIT_SYS;
+		goto err;
 	}
 
 	ret = ploop_get_size(mount_param.device, &dev_size);
 	if (ret)
 		goto err;
+	ret = ploop_get_size(part_device, &part_dev_size);
+	if (ret)
+		goto err;
+
+	if (new_size != 0) {
+		/* use (4 * blocksize) as reserved space for alignment */
+		if (new_size <= (4 * blocksize)) {
+			ploop_err(0, "Unable to change image size to %llu sectors",
+					new_size);
+			ret = SYSEXIT_PARAM;
+			goto err;
+		}
+		new_fs_size = new_size - (4 * blocksize);
+	}
 
 	ret = get_balloon(mount_param.target, &st, &balloonfd);
 	if (ret)
@@ -1554,13 +1740,26 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 		char conf[PATH_MAX];
 		char conf_tmp[PATH_MAX];
 
-		// GROW
+		/* GROW */
 		if (balloon_size != 0) {
 			ret = ploop_balloon_change_size(mount_param.device,
 					balloonfd, 0);
 			if (ret)
 				goto err;
 		}
+		close(balloonfd);
+		balloonfd = -1;
+		if (!mounted && param->offline_resize) {
+			/* offline */
+			if (do_umount(mount_param.target)) {
+				ploop_err(errno, "umount %s failed", mount_param.target);
+			} else {
+				ret = e2fsck(part_device);
+				if (ret)
+					goto err;
+			}
+		}
+
 		// Update size in the DiskDescriptor.xml
 		di->size = new_size;
 		get_disk_descriptor_fname(di, conf, sizeof(conf));
@@ -1569,7 +1768,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 		if (ret)
 			goto err;
 
-		ret = ploop_grow_device(mount_param.device, di->blocksize, new_size);
+		ret = ploop_grow_device(mount_param.device, blocksize, new_size);
 		if (ret) {
 			unlink(conf_tmp);
 			goto err;
@@ -1582,43 +1781,92 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			goto err;
 		}
 
-		ret = resize_fs(mount_param.device);
+		ret = resize_gpt_partition(mount_param.device, 0);
 		if (ret)
 			goto err;
-		tune_fs(mount_param.target, mount_param.device, new_size);
-	} else {
-		off_t available_balloon_size;
-		// SHRINK
-		/* FIXME: resize file system in case fs_size != dev_size
-		 */
-		if (statfs(mount_param.target, &fs) != 0) {
-			ploop_err(errno, "statfs(%s)", mount_param.target);
-			ret = SYSEXIT_FSTAT;
-			goto err;
-		}
-		new_balloon_size = dev_size - new_size;
-		available_balloon_size = balloon_size + B2S(fs.f_bfree * fs.f_bsize);
-		if (available_balloon_size < new_balloon_size) {
-			ploop_err(0, "Unable to change image size to %llu "
-					"sectors, minimal size is %lu",
-					new_size,
-					(long unsigned)(dev_size - available_balloon_size));
-			ret = SYSEXIT_PARAM;
-			goto err;
-		}
 
-		if (new_balloon_size != balloon_size) {
-			ret = ploop_balloon_change_size(mount_param.device,
-					balloonfd, new_balloon_size);
+		/* resize up to the end of device */
+		ret = resize_fs(part_device, new_fs_size);
+		if (ret)
+			goto err;
+	} else {
+		/* Grow or shrink fs but do not change block device size */
+		if (!mounted && param->offline_resize) {
+			/* Offline */
+			if (balloon_size != 0) {
+				/* FIXME: restore balloon size on failure */
+				ret = ploop_balloon_change_size(mount_param.device, balloonfd, 0);
+				if (ret)
+					goto err;
+			}
+			close(balloonfd); /* close to make umount possible */
+			balloonfd = -1;
+
+			ret = do_umount(mount_param.target);
+			if (ret) {
+				ploop_err(errno, "umount %s failed", mount_param.target);
+				ret = SYSEXIT_UMOUNT;
+				goto err;
+			}
+
+			ret = shrink_device(di, mount_param.device, part_device, part_dev_size,
+					new_fs_size, blocksize);
 			if (ret)
 				goto err;
-			tune_fs(mount_param.target, mount_param.device, new_size);
+		} else {
+			/* Online */
+			struct dump2fs_data data = {};
+			__u64 available_balloon_size;
+			__u64 blocks;
+
+			ret = dumpe2fs(part_device, &data);
+			if (ret)
+				goto err;
+
+			blocks = data.block_count * B2S(data.block_size);
+			if (new_fs_size < blocks) {
+				/* shrink fs */
+				if (statfs(mount_param.target, &fs) != 0) {
+					ploop_err(errno, "statfs(%s)", mount_param.target);
+					ret = SYSEXIT_FSTAT;
+					goto err;
+				}
+
+				new_balloon_size = blocks - new_fs_size;
+				available_balloon_size = balloon_size + (fs.f_bfree * B2S(fs.f_bsize));
+				if (available_balloon_size < new_balloon_size) {
+					ploop_err(0, "Unable to change image size to %lu "
+							"sectors, minimal size is %llu",
+							(long)new_fs_size,
+							(blocks - available_balloon_size));
+					ret = SYSEXIT_PARAM;
+					goto err;
+				}
+			} else {
+				/* grow fs */
+				new_balloon_size = 0;
+			}
+
+			if (new_balloon_size != balloon_size) {
+				ret = ploop_balloon_change_size(mount_param.device,
+						balloonfd, new_balloon_size);
+				if (ret)
+					goto err;
+				tune_fs(mount_param.target, part_device, new_fs_size);
+			}
+
+			if (new_balloon_size == 0) {
+				ret = resize_fs(part_device, new_fs_size);
+				if (ret)
+					goto err;
+			}
 		}
 	}
 
 err:
-	close(balloonfd);
-	if (mounted)
+	if (balloonfd != -1)
+		close(balloonfd);
+	if (mounted == 0)
 		ploop_umount(mount_param.device, di);
 	ploop_unlock_di(di);
 	free_mount_param(&mount_param);
@@ -1833,12 +2081,29 @@ static int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *
 
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
-	if (ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev)) == 0 &&
-			get_mount_dir(dev, mnt, sizeof(mnt)) == 0)
-	{
+
+	ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
+	if (ret == -1)
+		goto err;
+	if (ret == 0) {
+		ret = get_mount_dir(dev, mnt, sizeof(mnt));
+		if (ret)
+			goto err;
 		ret = get_statfs_info(mnt, info);
+	} else {
+		/* reinit .statfs */
+		struct ploop_mount_param param = {};
+
+		ret = auto_mount_image(di, &param);
+		if (ret == 0)
+			ploop_umount(param.device, di);
+		free_mount_param(&param);
+		ret = read_statfs_info(di->images[0]->file, info);
+		if (ret)
+			goto err;
 	}
 
+err:
 	ploop_unlock_di(di);
 
 	return ret;
@@ -1920,12 +2185,10 @@ int create_snapshot(const char *device, const char *delta, __u32 blocksize, int
 	req.c.pctl_cluster_log = ffs(blocksize) - 1;
 	req.c.pctl_size = 0;
 	req.c.pctl_chunks = 1;
-
 	req.f.pctl_type = PLOOP_IO_AUTO;
 
 	ploop_log(0, "Creating snapshot dev=%s img=%s", device, delta);
 	ret = do_snapshot(lfd, fd, &req);
-
 	if (ret)
 		unlink(delta);
 err:
@@ -2096,20 +2359,23 @@ err_cleanup1:
 	return ret;
 }
 
-int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, int flags)
+int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
+		struct ploop_snapshot_switch_param *param)
 {
 	int ret;
 	int fd;
 	char dev[64];
 	char uuid[61];
 	char file_uuid[61];
-	char new_top_delta_fname[PATH_MAX];
+	char new_top_delta_fname[PATH_MAX] = "";
 	char *old_top_delta_fname = NULL;
 	char conf[PATH_MAX];
 	char conf_tmp[PATH_MAX];
 	off_t size;
+	const char *guid = param->guid;
+	int flags = param->flags;
 
-	if (guid == NULL || !is_valid_guid(guid)) {
+	if (!is_valid_guid(guid)) {
 		ploop_err(0, "Incorrect guid %s", guid);
 		return SYSEXIT_PARAM;
 	}
@@ -2122,13 +2388,14 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		ploop_err(errno, "Nothing to do, already on %s snapshot",
 				guid);
 		goto err_cleanup1;
-
 	}
+
 	if (find_snapshot_by_guid(di, guid) == -1) {
 		ploop_err(0, "Can't find snapshot by uuid %s",
 				guid);
 		goto err_cleanup1;
 	}
+
 	// Read image size from image header
 	ret = get_image_size(di, guid, &size);
 	if (ret)
@@ -2140,10 +2407,6 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		goto err_cleanup1;
 	}
 
-	ret = ploop_di_remove_image(di, di->top_guid, 0, &old_top_delta_fname);
-	if (ret)
-		goto err_cleanup1;
-
 	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_DESTROY)) {
 		// device should be stopped
 		ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
@@ -2157,14 +2420,33 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 					dev);
 			goto err_cleanup1;
 		}
-	} else
-		old_top_delta_fname = NULL;
+		ret = ploop_di_remove_image(di, di->top_guid, 0, &old_top_delta_fname);
+		if (ret)
+			goto err_cleanup1;
+	} else if (param->guid_old != NULL) {
+		if (!is_valid_guid(param->guid_old)) {
+			ploop_err(0, "Incorrect guid %s", param->guid_old);
+			goto err_cleanup1;
+		}
 
-	snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
+		if (find_snapshot_by_guid(di, param->guid_old) != -1) {
+			ploop_err(0, "Incorrect guid_old %s: already exists",
+					param->guid_old);
+			goto err_cleanup1;
+		}
+
+		ploop_di_change_guid(di, di->top_guid, param->guid_old);
+	}
+
+	if (flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE) {
+		ploop_di_change_guid(di, guid, TOPDELTA_UUID);
+	} else {
+		snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
 			di->images[0]->file, file_uuid);
-	ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
-	if (ret)
-		goto err_cleanup1;
+		ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
+		if (ret)
+			goto err_cleanup1;
+	}
 
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
 	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
@@ -2173,12 +2455,14 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		goto err_cleanup1;
 
 	// offline snapshot
-	fd = create_empty_delta(new_top_delta_fname, di->blocksize, size);
-	if (fd == -1) {
-		ret = SYSEXIT_CREAT;
-		goto err_cleanup2;
+	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE)) {
+		fd = create_empty_delta(new_top_delta_fname, di->blocksize, size);
+		if (fd == -1) {
+			ret = SYSEXIT_CREAT;
+			goto err_cleanup2;
+		}
+		close(fd);
 	}
-	close(fd);
 
 	if (rename(conf_tmp, conf)) {
 		ploop_err(errno, "Can't rename %s %s",
@@ -2186,6 +2470,10 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 		ret = SYSEXIT_RENAME;
 		goto err_cleanup3;
 	}
+
+	/* destroy precached info */
+	drop_statfs_info(di->images[0]->file);
+
 	if (old_top_delta_fname != NULL) {
 		ploop_log(0, "Removing %s", old_top_delta_fname);
 		if (unlink(old_top_delta_fname))
@@ -2209,6 +2497,16 @@ err_cleanup1:
 	return ret;
 }
 
+int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, int flags)
+{
+	struct ploop_snapshot_switch_param param = {};
+
+	param.guid = (char *) guid;
+	param.flags = flags;
+
+	return ploop_switch_snapshot_ex(di, &param);
+}
+
 int ploop_delete_top_delta(struct ploop_disk_images_data *di)
 {
 	return ploop_delete_snapshot(di, di->top_guid);
