@@ -57,7 +57,11 @@
 
 #define S2B(sec) ((off_t)(sec) << PLOOP1_SECTOR_LOG)
 #define B2S(sec) ((sec) >> PLOOP1_SECTOR_LOG)
-#define ROUNDUP_BDSIZE(bdsize, blocksize) ((bdsize + blocksize - 1) / blocksize * blocksize)
+#define ROUNDUP_BDSIZE(size, blocksize) \
+	(((off_t)size + blocksize - 1) / blocksize * blocksize)
+#define ROUND_BDSIZE(size, blocksize) \
+	((size > 0xffffffffUL - blocksize) ? \
+	 (size / blocksize * blocksize) : ROUNDUP_BDSIZE(size, blocksize))
 
 struct delta_fops
 {
@@ -162,6 +166,12 @@ struct ploop_disk_images_runtime_data {
 	char *component_name;
 };
 
+struct dump2fs_data {
+	__u64 block_count;
+	__u64 block_free;
+	__u32 block_size;
+};
+
 enum {
 	PLOOP_MERGE_WITH_CHILD = 0,
 	PLOOP_MERGE_WITH_PARENT = 1,
@@ -171,26 +181,26 @@ enum {
 #define PL_EXT __attribute__ ((visibility("default")))
 
 int gen_uuid_pair(char *uuid1, int len1, char *uuid2, int len2);
-
 int find_delta_names(const char * device, int start_level, int end_level,
 			    char **names, char ** format);
 PL_EXT int ploop_get_attr(const char * device, const char * attr, int * res);
 int ploop_get_delta_attr(const char * device, int level, const char * attr, int * res);
 int ploop_get_size(const char * device, off_t * res);
 int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start);
+int get_dev_start(const char *path, __u32 *start);
 int ploop_get_top_level(int devfd, const char *devname, int *top);
 
 int init_delta_array(struct delta_array *);
 void deinit_delta_array(struct delta_array * p);
 int extend_delta_array(struct delta_array * p, char * path, int rw, int od_flags);
-PL_EXT void close_delta(struct delta *delta);
+void close_delta(struct delta *delta);
 PL_EXT int open_delta(struct delta * delta, const char * path, int rw, int od_flags);
-PL_EXT int open_delta_simple(struct delta * delta, const char * path, int rw, int od_flags);
+int open_delta_simple(struct delta * delta, const char * path, int rw, int od_flags);
 PL_EXT int dirty_delta(struct delta * delta);
 PL_EXT int clear_delta(struct delta * delta);
 PL_EXT int read_size_from_image(const char *img_name, int raw, off_t * res);
 PL_EXT int grow_delta(struct delta *odelta, off_t bdsize, void *buf,
-		       struct grow_maps *gm);
+		struct grow_maps *gm);
 PL_EXT int grow_raw_delta(const char *image, off_t append_size);
 PL_EXT int ploop_grow_device(const char *device, __u32 blocksize, off_t new_size);
 
@@ -270,8 +280,8 @@ int ploop_di_remove_image(struct ploop_disk_images_data *di, const char *guid,
 		int renew_top_uuid, char **fname);
 int ploop_di_merge_image(struct ploop_disk_images_data *di, const char *guid, char **fname);
 void ploop_di_change_guid(struct ploop_disk_images_data *di, const char *guid, const char *new_guid);
-char *find_image_by_guid(struct ploop_disk_images_data *di, const char *guid);
-int find_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid);
+PL_EXT char *find_image_by_guid(struct ploop_disk_images_data *di, const char *guid);
+PL_EXT int find_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_add_image_entry(struct ploop_disk_images_data *di, const char *fname, const char *guid);
 int ploop_add_snapshot_entry(struct ploop_disk_images_data *di, const char *guid,
 		const char *parent_guid);
@@ -289,6 +299,7 @@ PL_EXT int ploop_discard_get_stat_by_dev(const char *device, const char *mount_p
 		struct ploop_discard_stat *pd_stat);
 PL_EXT int ploop_discard_by_dev(const char *device, const char *mount_point,
 		__u64 minlen_b, __u64 to_free, const int *stop);
+int ploop_blk_discard(const char* device, __u32 blocksize, off_t start, off_t end);
 
 /* lock */
 int ploop_lock_di(struct ploop_disk_images_data *di);
@@ -300,13 +311,15 @@ void ploop_unlock(int *lckfd);
 int get_partition_device_name(const char *device, char *out, int size);
 int make_fs(const char *device, const char *fstype);
 void tune_fs(const char *target, const char *device, unsigned long long size);
-int resize_fs(const char *device);
+int resize_fs(const char *device, off_t blocks);
+int dumpe2fs(const char *device, struct dump2fs_data *data);
+int e2fsck(const char *device);
 int create_gpt_partition(const char *dev, off_t size, __u32 blocksize);
-int resize_gpt_partition(const char *devname);
+int resize_gpt_partition(const char *devname, __u64 new_size);
 
 // misc
 void get_basedir(const char *fname, char *out, int len);
-__u32 crc32(const unsigned char *buf, unsigned long len);
+__u32 ploop_crc32(const unsigned char *buf, unsigned long len);
 int store_statfs_info(const char *mnt, char *image);
 int drop_statfs_info(const char *image);
 int read_statfs_info(const char *image, struct ploop_info *info);
@@ -314,22 +327,20 @@ int get_statfs_info(const char *mnt, struct ploop_info *info);
 int ploop_get_child_count_by_uuid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_get_child_by_uuid(struct ploop_disk_images_data *di, const char *guid,  char **child_guid);
 int ploop_fname_cmp(const char *p1, const char *p2);
-int is_valid_guid(const char *guid);
+PL_EXT int is_valid_guid(const char *guid);
 PL_EXT int read_line(const char *path, char *nbuf, int len);
 int read_line_quiet(const char *path, char *nbuf, int len);
 int is_valid_blocksize(__u32 blocksize);
 int run_prg(char *const argv[]);
-int guidcmp(const char *p1, const char *p2);
+PL_EXT int guidcmp(const char *p1, const char *p2);
 int auto_mount_image(struct ploop_disk_images_data *di,
 		struct ploop_mount_param *param);
 void free_mount_param(struct ploop_mount_param *param);
 
-
 // merge
 PL_EXT int get_delta_info(const char *device, struct merge_info *info);
 PL_EXT int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
 		char **images);
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode);
-int ploop_complete_running_operation(const char *device);
 
 #endif
