@@ -29,13 +29,14 @@
 #include <string.h>
 
 #include "ploop.h"
+#include "common.h"
 
-int raw;
-struct delta delta;
-struct ploop_pvd_header new_vh;
-char *device;
+static int raw;
+static struct delta delta;
+static struct ploop_pvd_header new_vh;
+static char *device;
 
-void usage(char *pname)
+static void usage(char *pname)
 {
 	fprintf(stderr, "Usage: %s [-s NEW_SIZE] -d DEVICE\n"
 		        "       %s [-s NEW_SIZE] [-f raw] DELTA\n",
@@ -129,18 +130,18 @@ int
 main(int argc, char ** argv)
 {
 	char *pname = argv[0];
-	int i;
+	int i, f;
 	off_t new_size = 0; /* in sectors */
 
 	while ((i = getopt(argc, argv, "f:d:s:")) != EOF) {
 		switch (i) {
 		case 'f':
-			if (strcmp(optarg, "raw") == 0)
-				raw = 1;
-			else if (strcmp(optarg, "direct") != 0) {
+			f = parse_format_opt(optarg);
+			if (f < 0) {
 				usage(pname);
 				return -1;
 			}
+			raw = (f == PLOOP_RAW_MODE);
 			break;
 		case 'd':
 			device = optarg;
