@@ -33,6 +33,8 @@
 #include "ploop.h"
 #include "common.h"
 
+extern int plooptool_snapshot_list(int argc, char **argv);
+
 static void usage_summary(void)
 {
 	fprintf(stderr, "Usage: ploop init -s SIZE [-f FORMAT] NEW_DELTA\n"
@@ -48,8 +50,9 @@ static void usage_summary(void)
 			"       ploop snapshot-delete -u <uuid> DiskDescriptor.xml\n"
 			"       ploop snapshot-merge [-u <uuid>] DiskDescriptor.xml\n"
 			"       ploop snapshot-switch -u <uuid> DiskDescriptor.xml\n"
+			"       ploop snapshot-list [-o field[,field...]] [--uuid <UUID>] DiskDescriptor.xml\n"
 			"Also:  ploop { stat | start | stop | clear } ...\n"
-	       );
+		);
 }
 
 static void usage_init(void)
@@ -502,7 +505,7 @@ static int plooptool_rm(int argc, char **argv)
 
 	lfd = open(rmopts.device, O_RDONLY);
 	if (lfd < 0) {
-		perror("open dev");
+		perror("Can't open device");
 		return SYSEXIT_DEVICE;
 	}
 
@@ -755,7 +758,7 @@ static int plooptool_resize(int argc, char **argv)
 	int i, ret;
 	off_t new_size = 0; /* in sectors */
 	int max_balloon_size = 0; /* make balloon file of max possible size */
-	struct ploop_resize_param param = {};
+	struct ploop_resize_param param = {.size = 0, .offline_resize = 1};
 	struct ploop_disk_images_data *di;
 
 	while ((i = getopt(argc, argv, "s:b")) != EOF) {
@@ -881,14 +884,34 @@ static int plooptool_info(int argc, char **argv)
 	return ret;
 }
 
+static void usage_list(void)
+{
+	fprintf(stderr, "Usage: ploop list [-a]\n");
+}
+
 static int plooptool_list(int argc, char **argv)
 {
 	char fname[PATH_MAX];
 	char image[PATH_MAX];
+	char mnt[PATH_MAX] = "";
+	char dev[64];
 	DIR *dp;
 	struct stat st;
 	struct dirent *de;
 	char cookie[PLOOP_COOKIE_SIZE];
+	int all = 0;
+	int i;
+
+	while ((i = getopt(argc, argv, "a")) != EOF) {
+		switch (i) {
+		case 'a':
+			all = 1;
+			break;
+		default:
+			usage_list();
+			return SYSEXIT_PARAM;
+		}
+	}
 
 	snprintf(fname, sizeof(fname) - 1, "/sys/block/");
 	dp = opendir(fname);
@@ -913,30 +936,18 @@ static int plooptool_list(int argc, char **argv)
 				continue;
 		}
 
-		printf("%-12s %s %s\n", de->d_name, image, cookie);
+		if (all) {
+			mnt[0] = '\0';
+			snprintf(dev, sizeof(dev), "/dev/%s", de->d_name);
+			ploop_get_mnt_by_dev(dev, mnt, sizeof(mnt));
+		}
+		printf("%-12s %s %s %s\n", de->d_name, image, mnt, cookie);
 	}
 	closedir(dp);
 
 	return 0;
 }
 
-
-static void cancel_callback(int sig)
-{
-	ploop_cancel_operation();
-}
-
-static void init_signals(void)
-{
-	struct sigaction act = {};
-
-	sigemptyset(&act.sa_mask);
-	act.sa_handler = cancel_callback;
-	sigaction(SIGTERM, &act, NULL);
-	sigaction(SIGINT, &act, NULL);
-	sigaction(SIGHUP, &act, NULL);
-}
-
 int main(int argc, char **argv)
 {
 	char * cmd;
@@ -975,6 +986,8 @@ int main(int argc, char **argv)
 		return plooptool_snapshot_delete(argc, argv);
 	if (strcmp(cmd, "snapshot-merge") == 0)
 		return plooptool_snapshot_merge(argc, argv);
+	if (strcmp(cmd, "snapshot-list") == 0)
+		return plooptool_snapshot_list(argc, argv);
 	if (strcmp(cmd, "getdev") == 0)
 		return plooptool_getdevice(argc, argv);
 	if (strcmp(cmd, "resize") == 0)
